/* hex2raw.c			letzte nderung: 15.9.2012 */
static char *VERSION="0.0";
/*
Umwandlung vom hex-Format in Rohdaten
Programmiert nach den Angaben von dieser Webseite:
 http://de.wikipedia.org/wiki/Intel_HEX

History:
Erstellung:	15.9.2012 (RPf)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#define MAXARG 2

/************************ Vordeklarationen *******************************/
void hex2raw(FILE *,FILE *);
/*************************************************************************/

void usage()
{
 printf("hex2raw Version %s\n",VERSION);
 printf("Anwendung: hex2raw Name.hex Name.raw [-Flags]\n");
 printf("  Flags:\n");
 printf("      h = Hilfe\n");
 exit(0);
}

static char argflag['Z'+1];

void setflags(char *s)
{
 int c;
 while((c= *s++)!=0)
  if((c=toupper(c))<='Z')  argflag[c]=1;
}

int main(int argc,char *argv[])
{
 FILE *fp,*fp2;
 int i,j;
 char name1[80],name2[80],*rstring,*wstring;
 for(name1[0]=name2[0]=0,i=j=1;i<argc;i++)
   if(*argv[i]=='-') setflags(argv[i]);
   else switch(j++)
        {case 1:strcpy(name1,argv[i]); break;
         case 2:strcpy(name2,argv[i]); break;
        }
 if(j>MAXARG+1 || name1[0]=='?' || argflag['?'] || argflag['H']) usage();
 if(name1[0]==0) {printf("Hexdatei:"); i=scanf("%s",name1);}
 if(name2[0]==0) {printf("Zieldatei:"); i=scanf("%s",name2);}
 if(argflag['X']) {rstring="r"; wstring="w";}
 else		 {rstring="rb"; wstring="wb";}
 if(!(fp=fopen(name1,rstring)))
	{printf("'%s' nicht gefunden\n",name1); exit(0);}
 if(!(fp2=fopen(name2,wstring)))
	{printf("kann '%s' nicht ffnen\n",name2); fclose(fp); exit(0);}
 hex2raw(fp,fp2);
 fclose(fp); fclose(fp2);
 return 0;
}

/*************************** kleinkram ***********************************/
int ascitodigit(int c)
{		/* wandelt ein ASCI-Zeichen in eine 1-stellige Hex-Zahl */
 if(!isxdigit(c)) return 0;
 if(islower(c)) c=toupper(c)-'0';
 else c-='0';
 if(c>9) c += '0'-'A'+10;
 return c;
}

#define getline mygetline

int getline(FILE *fp,char *s,int lim)
{		/* liest eine Textzeile oder maximal lim Zeichen */
		/* und ersetzt den Zeilentrenner durch 0         */
 int c=EOF;
 while(--lim && (c=getc(fp))!=EOF && c!='\n')
    {if(c!='\r')  *s++ = c;}
 *s='\0';
 return (c!=EOF);	/* TRUE wenn erfolgreich, FALSE wenn Fileende */
}

void putline(FILE *fp,const char *s) /* wie fprintf(fp,"%s\n",s)  aber */
{				     /* ohne ein Nullbyte zu schreiben */
 char c;
 while((c= *s++)!=0)  putc(c,fp);
 putc('\n',fp);
}

/*************************** Hauptteil ***********************************/
int pruefsummecheck(const char *s)
{
 int pruefsumme=0,c1,c2;
 while((c1= *s++)!=0 && (c2= *s++)!=0)
  {
   pruefsumme += (ascitodigit(c1)<<4)+ascitodigit(c2);
  }
 return (pruefsumme&0xFF);
}

void hex2raw(FILE *fp,FILE *fp2)
{
 int c,i,j,reclen,rectyp;
 long adresse= -1, extadresse=0, adr;
 char zeile[120];
 while(getline(fp,zeile,120))
  {
   if(zeile[0]!=':') {printf("Fehler: unbekannte Zeile '%s'\n",zeile); continue;}
   if(pruefsummecheck(&zeile[1])!=0) {printf(" Fehler: Pruefsumme falsch in Zeile '%s'\n",zeile); return;}
   j=1;
   reclen = (ascitodigit(zeile[j])<<4) + ascitodigit(zeile[j+1]);
   j+=2;
   adr = (ascitodigit(zeile[j])<<12)+(ascitodigit(zeile[j+1])<<8)+(ascitodigit(zeile[j+2])<<4)+ascitodigit(zeile[j+3]);
   j+=4;
   rectyp = (ascitodigit(zeile[j])<<4) + ascitodigit(zeile[j+1]);
   j+=2;
   if(rectyp==0) //Data Record, Nutzdaten
    {
     if(adresse<0) adresse=adr;
     else if(adr>adresse)
      {
       printf("Warnung: Luecke im Datensatz zwischen %ld und %ld - wird mit 0xFF aufgefuellt.\n",adresse,adr);
       while(adresse<adr) {putc(0xFF,fp2); adresse++;}
      }
     else if(adr<adresse)
      {
       printf("Fehler: Adresse neu gesetzt von %ld auf %ld\n",adresse,adr);
       adresse=adr;
      }
     for(i=0;i<reclen;i++)
      {
       c = (ascitodigit(zeile[j])<<4) + ascitodigit(zeile[j+1]);
       j+=2;
       if(++adresse==0x10000) {adresse=0; extadresse+=0x10000;}
       putc(c,fp2);
      }
    }
   else if(rectyp==1) //End of File
    {
     printf("End of File Record - Umwandlung abgeschlossen.\n");
     return;
    }
   else if(rectyp==2) //Extended Segment Address Record
    {
     extadresse = (ascitodigit(zeile[j])<<12)+(ascitodigit(zeile[j+1])<<8)+(ascitodigit(zeile[j+2])<<4)+ascitodigit(zeile[j+3]);
     extadresse *= 16;
     printf("Test: extadresse = %ld = 0x%lX\n",extadresse,extadresse); //test
    }
   else if(rectyp==3) //Start Segment Address Record
    {
     printf("Start Segment Address Record - noch nicht unterstuetzt.\n");
    }
   else if(rectyp==4) //Extended Linear Address Record
    {
     printf("Extended Linear Address Record - noch nicht unterstuetzt.\n");
    }
   else if(rectyp==5) //Start Linear Address Record
    {
     printf("Start Linear Address Record - noch nicht unterstuetzt.\n");
    }
  }
}
